@@ -22,17 +22,16 @@ from mo_dots import (
     get_attr,
     listwrap,
     unwraplist,
-    dict_to_data, Data, join_field,
+    dict_to_data,
 )
 from mo_files import File
 from mo_files.url import URL
-from mo_future import is_text
-from mo_future import text
 from mo_json import json2value
-from mo_logs import Except, logger
+from mo_logs import Except, logger, get_stacktrace
 
 from mo_json_config.configuration import Configuration
 from mo_json_config.convert import ini2value
+from mo_json_config.ssm import get_ssm as _get_ssm
 
 DEBUG = False
 
@@ -46,25 +45,34 @@ def get(url):
     """
     USE json.net CONVENTIONS TO LINK TO INLINE OTHER JSON
     """
-    url = text(url)
+    url = str(url)
     if "://" not in url:
         logger.error("{{url}} must have a prototcol (eg http://) declared", url=url)
+    path = (dict_to_data({"$ref": url}), None)
 
-    base = URL("")
     if url.startswith("file://") and url[7] != "/":
-        if os.sep == "\\":
-            base = URL("file:///" + os.getcwd().replace(os.sep, "/").rstrip("/") + "/.")
+        causes= []
+        candidates = [os.path.dirname(os.path.abspath(get_stacktrace(start=1)[0]['file'])), os.getcwd()]
+        for candidate in candidates:
+            if os.sep == "\\":
+                base = URL("file:///" + candidate.replace(os.sep, "/").rstrip("/") + "/.")
+            else:
+                base = URL("file://" + candidate.rstrip("/") + "/.")
+            try:
+                phase1 = _replace_ref(path, base)
+                break
+            except Exception as cause:
+                causes.append(cause)
         else:
-            base = URL("file://" + os.getcwd().rstrip("/") + "/.")
+            logger.error("problem replacing ref in {{url}}", url=url, cause=causes)
+    else:
+        phase1 = _replace_ref(path, URL(""))  # BLANK URL ONLY WORKS IF url IS ABSOLUTE
 
-    phase1 = _replace_ref(
-        dict_to_data({"$ref": url}), base
-    )  # BLANK URL ONLY WORKS IF url IS ABSOLUTE
     try:
-        phase2 = _replace_locals(phase1, [phase1])
+        phase2 = _replace_locals((phase1,None), url)
         return to_data(phase2)
-    except Exception as e:
-        logger.error("problem replacing locals in\n{{phase1}}", phase1=phase1, cause=e)
+    except Exception as cause:
+        logger.error("problem replacing locals in\n{{phase1}}", phase1=phase1, cause=cause)
 
 
 def expand(doc, doc_url="param://", params=None):
@@ -80,27 +88,52 @@ def expand(doc, doc_url="param://", params=None):
     :return: EXPANDED JSON-SERIALIZABLE STRUCTURE
     """
     if "://" not in doc_url:
-        logger.error("{{url}} must have a prototcol (eg http://) declared", url=doc_url)
+        logger.error("{{url}} must have a protocol (eg https://) declared", url=doc_url)
 
     url = URL(doc_url)
     url.query = set_default(url.query, params)
-    phase1 = _replace_ref(doc, url)  # BLANK URL ONLY WORKS IF url IS ABSOLUTE
-    phase2 = _replace_locals(phase1, [phase1])
+    phase1 = _replace_ref((doc, None), url)  # BLANK URL ONLY WORKS IF url IS ABSOLUTE
+    phase2 = _replace_locals((phase1,None), url)
     return to_data(phase2)
 
 
-def _replace_ref(node, url):
+is_url = re.compile(r"\{([0-9a-zA-Z]+://[^}]*)}")
+
+
+def _replace_str(text, path, url):
+    acc = []
+    end = 0
+    for found in is_url.finditer(text):
+        acc.append(text[end: found.start()])
+        try:
+            ref = URL(found.group(1))
+            if ref.scheme not in scheme_loaders:
+                raise logger.error("unknown protocol {ref}", ref=ref)
+            value = scheme_loaders[ref.scheme](ref, path, url)
+            if value == None:
+                raise logger.error("value not found {ref}", ref=ref)
+            acc.append(value)
+        except Exception as cause:
+            raise logger.error("problem replacing {ref}", ref=found.group(1), cause=cause)
+        end = found.end()
+    if end == 0:
+        return text
+    return "".join(acc) + text[end:]
+
+
+def _replace_ref(path, url):
     if url.path.endswith("/"):
         url.path = url.path[:-1]
 
+    node = path[0]
     if is_data(node):
         refs = None
         output = {}
         for k, v in node.items():
             if k == "$ref":
-                refs = URL(v)
+                refs = URL(_replace_str(str(v), path, url))
             else:
-                output[k] = _replace_ref(v, url)
+                output[k] = _replace_ref((v, path), url)
 
         if not refs:
             return output
@@ -124,23 +157,20 @@ def _replace_ref(node, url):
             if ref.scheme not in scheme_loaders:
                 raise logger.error("unknown protocol {{scheme}}", scheme=ref.scheme)
             try:
-                new_value = scheme_loaders[ref.scheme](ref, url)
+                new_value = scheme_loaders[ref.scheme](ref, (node, path), url)
                 ref_found = True
-            except Exception as e:
-                e = Except.wrap(e)
-                ref_error = e
+            except Exception as cause:
+                ref_error = Except.wrap(cause)
                 continue
 
             if ref.fragment:
                 new_value = get_attr(new_value, ref.fragment)
 
-            DEBUG and logger.note(
-                "Replace {{ref}} with {{new_value}}", ref=ref, new_value=new_value
-            )
+            DEBUG and logger.note("Replace {{ref}} with {{new_value}}", ref=ref, new_value=new_value)
 
             if not output:
                 output = new_value
-            elif is_text(output):
+            elif isinstance(output, str):
                 pass  # WE HAVE A VALUE
             else:
                 set_default(output, new_value)
@@ -152,54 +182,34 @@ def _replace_ref(node, url):
         DEBUG and logger.note("Return {{output}}", output=output)
         return output
     elif is_list(node):
-        output = [_replace_ref(n, url) for n in node]
-        # if all(p[0] is p[1] for p in zip(output, node)):
-        #     return node
+        output = [_replace_ref((n, path), url) for n in node]
         return output
 
     return node
 
 
-def _replace_locals(node, doc_path):
+def _replace_locals(path, url):
+    node = path[0]
     if is_data(node):
         # RECURS, DEEP COPY
         ref = None
         output = {}
         for k, v in node.items():
             if k == "$ref":
-                ref = v
+                ref = URL(_replace_str(str(v), path, url))
             elif k == "$concat":
                 if not is_sequence(v):
                     logger.error("$concat expects an array of strings")
-                return coalesce(node.get("separator"), "").join(v)
+                return coalesce(node.get("separator"), "").join(_replace_locals((vv, path), url) for vv in v)
             elif v == None:
                 continue
             else:
-                output[k] = _replace_locals(v, [v] + doc_path)
+                output[k] = _replace_locals((v, path), url)
 
         if not ref:
             return output
 
-        # REFER TO SELF
-        frag = ref.fragment
-        if frag[0] == ".":
-            # RELATIVE
-            for i, p in enumerate(frag):
-                if p != ".":
-                    if i > len(doc_path):
-                        logger.error(
-                            "{{frag|quote}} reaches up past the root document",
-                            frag=frag,
-                        )
-                    new_value = get_attr(doc_path[i - 1], frag[i::])
-                    break
-            else:
-                new_value = doc_path[len(frag) - 1]
-        else:
-            # ABSOLUTE
-            new_value = get_attr(doc_path[-1], frag)
-
-        new_value = _replace_locals(new_value, [new_value] + doc_path)
+        new_value = _get_value_from_fragment(ref, path, url)
 
         if not output:
             return new_value  # OPTIMIZATION FOR CASE WHEN node IS {}
@@ -207,20 +217,43 @@ def _replace_locals(node, doc_path):
             return from_data(set_default(output, new_value))
 
     elif is_list(node):
-        candidate = [_replace_locals(n, [n] + doc_path) for n in node]
-        # if all(p[0] is p[1] for p in zip(candidate, node)):
-        #     return node
-        return candidate
+        return [_replace_locals((n, path), url) for n in node]
 
+    elif isinstance(node, str):
+        return _replace_str(node, path[1], url)
     return node
 
 
+def _get_value_from_fragment(ref, path, url):
+    # REFER TO SELF
+    frag = ref.fragment
+    if frag[0] == ".":
+        doc = (None, path)
+        # RELATIVE
+        for i, c in enumerate(frag):
+            if c == ".":
+                if not isinstance(doc, tuple):
+                    logger.error("{frag|quote} reaches up past the root document", frag=frag)
+                doc = doc[1]
+            else:
+                break
+        new_value = get_attr(doc[0], frag[i::])
+    else:
+        # ABSOLUTE
+        top_doc = path
+        while isinstance(top_doc, tuple) and top_doc[1]:
+            top_doc = top_doc[1]
+        new_value = get_attr(top_doc[0], frag)
+    new_value = _replace_locals((new_value, path), url)
+    return new_value
+
+
 ###############################################################################
 ## SCHEME LOADERS ARE BELOW THIS LINE
 ###############################################################################
 
 
-def _get_file(ref, url):
+def _get_file(ref, path, url):
 
     if ref.path.startswith("~"):
         home_path = os.path.expanduser("~")
@@ -250,7 +283,7 @@ def _get_file(ref, url):
         content = File(path).read()
     except Exception as e:
         content = None
-        logger.error("Could not read file {{filename}}", filename=path, cause=e)
+        logger.error("Could not read file {{filename}}", filename=File(path).os_path, cause=e)
 
     try:
         new_value = json2value(content, params=ref.query, flexible=True, leaves=True)
@@ -260,19 +293,19 @@ def _get_file(ref, url):
             new_value = ini2value(content)
         except Exception:
             raise logger.error("Can not read {{file}}", file=path, cause=e)
-    new_value = _replace_ref(new_value, ref)
+    new_value = _replace_ref((new_value, path), ref)
     return new_value
 
 
-def get_http(ref, url):
+def get_http(ref, doc_path, url):
     import requests
 
     params = url.query
-    new_value = json2value(requests.get(str(ref)).json(), params=params, flexible=True, leaves=True)
+    new_value = json2value(requests.get(str(ref)).text, params=params, flexible=True, leaves=True)
     return new_value
 
 
-def _get_env(ref, url):
+def _get_env(ref, doc_path, url):
     # GET ENVIRONMENT VARIABLES
     ref = ref.host
     raw_value = os.environ.get(ref)
@@ -286,7 +319,7 @@ def _get_env(ref, url):
     return new_value
 
 
-def _get_keyring(ref, url):
+def _get_keyring(ref, doc_path, url):
     try:
         import keyring
     except Exception:
@@ -302,8 +335,7 @@ def _get_keyring(ref, url):
     raw_value = keyring.get_password(service_name, username)
     if not raw_value:
         logger.error(
-            "expecting password in the keyring for service_name={{service_name}} and"
-            " username={{username}}",
+            "expecting password in the keyring for service_name={{service_name}} and username={{username}}",
             service_name=service_name,
             username=username,
         )
@@ -315,38 +347,7 @@ def _get_keyring(ref, url):
     return new_value
 
 
-def _get_ssm(ref, url):
-    output = Data()
-    try:
-        import boto3
-    except Exception:
-        logger.error("Missing boto3: `pip install boto3` to use ssm://")
-    try:
-        ssm = boto3.client('ssm', region_name='us-east-1')
-        result = ssm.describe_parameters(MaxResults=10)
-        prefix = re.compile("^"+re.escape(ref.path.rstrip("/"))+"/|$")
-        while True:
-            for param in result['Parameters']:
-                name = param['Name']
-                found = prefix.match(name)
-                if not found:
-                    continue
-                tail = join_field(name[found.regs[0][1]:].split("/"))
-                detail = ssm.get_parameter(Name=name, WithDecryption=True)
-                output[tail] = detail['Parameter']['Value']
-
-            next_token = result.get('NextToken')
-            if not next_token:
-                break
-            result = ssm.describe_parameters(NextToken=next_token, MaxResults=10)
-    except Exception as cause:
-        logger.warning("Could not get ssm parameters", cause=cause)
-    if len(output) == 0:
-        logger.error("No ssm parameters found at {{path}}", path=ref.path)
-    return output
-
-
-def _get_param(ref, url):
+def _get_param(ref, doc_path, url):
     # GET PARAMETERS FROM url
     param = url.query
     new_value = param[ref.host]
@@ -361,6 +362,7 @@ scheme_loaders = {
     "param": _get_param,
     "keyring": _get_keyring,
     "ssm": _get_ssm,
+    "ref": _get_value_from_fragment,
 }
 
 
